//============================================================================
// Name        : levelSet.cpp
// Author      : Ojaswa
// Version     :
// Copyright   : Your copyright notice
// Description : Hello World in C, Ansi-style
//============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <cutil.h>
#include <cuda.h>

#define BLOCK_DIM 4
#define SUBVOL_DIM 64

enum {BBOX, SUPER_ELLIPSOID};

#include "global.cu"
#include "init_surface.cu"

#define CHK_CUDA_ERR(_str)    cuda_err = cudaGetLastError(); \
if(cuda_err != cudaSuccess) { \
fprintf(stderr,"%s: %s\n", _str, cudaGetErrorString(cuda_err)) ; \
return(false); \
} \
else \
fprintf(stderr,"%s: %s\n", _str, cudaGetErrorString(cuda_err)) ;

int iDivUp(int a, int b)
{
  return ((a % b) != 0)? (a / b + 1): (a / b);
}

const cudaExtent volSize = make_cudaExtent(1*(SUBVOL_DIM - 2) + 2, 1*(SUBVOL_DIM - 2) + 2, 1*(SUBVOL_DIM - 2) + 2); // Test size for a large volume that can encomapss 8x4x2 subvolumes
const cudaExtent subvolSize = make_cudaExtent(SUBVOL_DIM, SUBVOL_DIM, SUBVOL_DIM);
const cudaExtent logicalGridSize = make_cudaExtent(iDivUp(subvolSize.width, BLOCK_DIM), iDivUp(subvolSize.height, BLOCK_DIM), iDivUp(subvolSize.depth, BLOCK_DIM));
const cudaExtent subvolIndicesExtents = make_cudaExtent(iDivUp(volSize.width-2, subvolSize.width-2), iDivUp(volSize.height-2, subvolSize.height-2), iDivUp(volSize.depth-2, subvolSize.depth-2));

float* h_phi = NULL; // The full volume 
float* h_subvol = NULL; // subvolume SUBVOL_DIM^3
cudaPitchedPtr d_volPPtr;
float* d_subvol;
size_t tPitch = 0;
unsigned int timer = 0;
int init_interface_method = BBOX;
cudaError_t cuda_err;
cudaArray *d_phiArray = 0; // subvolume bound to texture - used for texture reads in phi
cudaArray *d_volArray = 0; // subvolume bound to texture - used for texture reads in intensitintensity

int BBoxOffset = 5;
float superEllipsoidPower = 4.0;
int subvolDim = SUBVOL_DIM;

const dim3 blockSize(BLOCK_DIM, BLOCK_DIM, BLOCK_DIM);
const dim3 gridSize(logicalGridSize.width*logicalGridSize.height*logicalGridSize.depth, 1);

// Host - host mem, Mem - device mem, Array - device array (texture bound)
void copy3DHostToArray(float *_src, cudaArray *_dst, cudaExtent copy_extent, cudaPos src_offset)
{
  cudaMemcpy3DParms copyParams = {0};
  float *h_source = _src + src_offset.x + src_offset.y*volSize.width + src_offset.z*volSize.width*volSize.height;
  copyParams.srcPtr = make_cudaPitchedPtr((void*)h_source, volSize.width*sizeof(float), volSize.width, volSize.height);
  copyParams.dstArray = _dst;
  copyParams.kind = cudaMemcpyHostToDevice;
  //copyParams.extent = make_cudaExtent(subvolDim, subvolDim, subvolDim);
  copyParams.extent = copy_extent;

  CUDA_SAFE_CALL(cudaMemcpy3D(&copyParams));
  CUT_CHECK_ERROR("Host -> Array Memcpy failed\n");
}

void copy3DMemToArray(cudaPitchedPtr _src, cudaArray *_dst)
{
  cudaMemcpy3DParms copyParams = {0};
  copyParams.srcPtr = 	_src;
  copyParams.dstArray = _dst;
  copyParams.kind = cudaMemcpyDeviceToDevice;
  copyParams.extent = make_cudaExtent(subvolDim, subvolDim, subvolDim);

  CUDA_SAFE_CALL(cudaMemcpy3D(&copyParams));
  CUT_CHECK_ERROR("Mem -> Array Memcpy failed\n");
}

void copy3DMemToHost(cudaPitchedPtr _src, float *_dst, cudaExtent copy_extent, cudaExtent dst_extent, cudaPos src_offset, cudaPos dst_offset)
{
  cudaMemcpy3DParms copyParams = {0};
  copyParams.srcPtr = _src;
  float *h_target = _dst + dst_offset.x + dst_offset.y*dst_extent.width + dst_offset.z*dst_extent.width*dst_extent.height;//For some reason, using copyParams.dstPos doesn't give correct results, so we set the offset here.
  copyParams.dstPtr = make_cudaPitchedPtr((void*)h_target, dst_extent.width*sizeof(float), dst_extent.width, dst_extent.height);
  copyParams.kind = cudaMemcpyDeviceToHost;
  copyParams.extent = make_cudaExtent(copy_extent.width*sizeof(float), copy_extent.height, copy_extent.depth);
  copyParams.srcPos = make_cudaPos(src_offset.x*sizeof(float), src_offset.y, src_offset.z); // We want to copy copy_extent sized volume starting at (x_off, y_off, z_off).

  CUDA_SAFE_CALL(cudaMemcpy3D(&copyParams));
  CUT_CHECK_ERROR("Mem -> Host Memcpy failed\n");
}

bool initCuda()
{
  //create 3D array
  cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<float>();
  cudaExtent subvolSize = make_cudaExtent(subvolDim, subvolDim, subvolDim);

  //Allocate mem and bind phiArray to 3D texture
  CUDA_SAFE_CALL(cudaMalloc3DArray(&d_phiArray, &channelDesc, subvolSize));
  CHK_CUDA_ERR("\n");
  texPhi.normalized = false;
  texPhi.filterMode = cudaFilterModePoint;
  texPhi.addressMode[0] = cudaAddressModeWrap; 
  texPhi.addressMode[1] = cudaAddressModeWrap; 
  texPhi.addressMode[2] = cudaAddressModeWrap; 
  CUDA_SAFE_CALL(cudaBindTextureToArray(texPhi, d_phiArray, channelDesc));
  CHK_CUDA_ERR("\n");

  //Allocate memory for 3D device memory for kernel output
  cudaExtent pitchedVolSize = make_cudaExtent(subvolDim*sizeof(float), subvolDim, subvolDim); 
  CUDA_SAFE_CALL(cudaMalloc3D(&d_volPPtr, pitchedVolSize));
  CHK_CUDA_ERR("\n");

  return true;
}

float* loadRawData()
{
  size_t sz = subvolSize.width*subvolSize.height*subvolSize.depth*sizeof(float);
  float *data = (float *)malloc(sz);
  memset(data, 0, sz);
  return data;
}

void writeRawData(float *data, char* filename)
{
  FILE *fp = fopen(filename, "w");
  if(!fp) {
    fprintf(stderr, "Error opening file '%s'\n", filename);
    return;
  }

  for(int _z = 0; _z<volSize.depth; _z++)
  {
    for(int _y = 0; _y<volSize.height; _y++)
    {
      for(int _x = 0; _x<volSize.width; _x++)
        fprintf(fp, "%+03.1f ", *(data + _x + _y*volSize.width + _z*volSize.width*volSize.height));
      fprintf(fp, "\n");
    }
    fprintf(fp, "--(%d)--\n", _z+1);
  }
  fclose(fp);
}

bool initInterface()
{
  /// Initialize level set interface (to a cuboid)
  fprintf(stderr, "Initializing the level set interface... ");
  float maxDist = 1.0; 
  //float maxDist = sqrt(volSize.width*volSize.width + volSize.height*volSize.height + volSize.depth*volSize.depth);
  char _msg[256];
  
  //The 1 voxel border is not initialized by our cuda kernel. Do it on the CPU :p. Outside is -ve
  for(int _x=0;_x<volSize.width; _x++) for(int _y=0; _y<volSize.height; _y++) { *(h_phi+ _x + _y*volSize.width) = -maxDist; *(h_phi + _x + _y*volSize.width + (volSize.depth-1)*volSize.width*volSize.height) = -maxDist; }
  for(int _x=0;_x<volSize.width; _x++) for(int _z=0; _z<volSize.depth; _z++) { *(h_phi+ _x + _z*volSize.width*volSize.height) = -maxDist; *(h_phi + _x + (volSize.height-1)*volSize.width + _z*volSize.width*volSize.height) = -maxDist; }
  for(int _z=0;_z<volSize.depth; _z++) for(int _y=0; _y<volSize.height; _y++) { *(h_phi + _y*volSize.width + _z*volSize.width*volSize.height) = -maxDist; *(h_phi + volSize.width - 1 + _y*volSize.width + _z*volSize.width*volSize.height) = -maxDist; }

  CUT_SAFE_CALL(cutStartTimer(timer));
  unsigned int planes[] = {BBoxOffset, volSize.width-BBoxOffset, BBoxOffset, volSize.height-BBoxOffset, BBoxOffset, volSize.depth-BBoxOffset};
  cudaExtent subvolIdx = make_cudaExtent(0, 0, 0);

  cudaExtent copyvol = make_cudaExtent(subvolDim-2, subvolDim-2, subvolDim-2);
  cudaPos src_off = make_cudaPos(1, 1, 1);
  cudaPos dst_off = make_cudaPos(0, 0, 0);

  cudaExtent subvolIndicesExtents;

  subvolIndicesExtents = make_cudaExtent(iDivUp(volSize.width-2, subvolDim-2), iDivUp(volSize.height-2, subvolDim-2), iDivUp(volSize.depth-2, subvolDim-2));
  for(int _z = 0; _z< subvolIndicesExtents.depth; _z++)
    for(int _y = 0; _y< subvolIndicesExtents.height; _y++)
      for(int _x = 0; _x< subvolIndicesExtents.width; _x++)
      {
  	cudaMemset3D(d_volPPtr, 0, make_cudaExtent(subvolDim*sizeof(float), subvolDim, subvolDim));// zero out the device mem
	subvolIdx.width = _x; subvolIdx.height = _y; subvolIdx.depth = _z;
	if(init_interface_method == BBOX)
     	  d_init_interface_BBox<<<gridSize, blockSize>>>(d_volPPtr, logicalGridSize, planes[0], planes[1], planes[2], planes[3], planes[4], planes[5], subvolIdx, maxDist, subvolDim, BLOCK_DIM);	
  	else if (init_interface_method == SUPER_ELLIPSOID)//superEllipsoidal interface
    	  d_init_interface_superEllipsoid<<<gridSize, blockSize>>>(d_volPPtr, logicalGridSize, planes[0], planes[1], planes[2], planes[3], planes[4], planes[5], subvolIdx, maxDist, subvolDim, BLOCK_DIM, superEllipsoidPower);	
    	else
	{
	  fprintf(stderr, "No interface initialization method specified. Exiting...\n");
	  return false;
	}
  	CUT_CHECK_ERROR("Kernel failed");
     	sprintf(_msg,  "---%d-%d-%d", _x, _y, _z);
     	CHK_CUDA_ERR(_msg);
  	cudaThreadSynchronize();

	//copy the results to main volume
	dst_off.x = 1 + _x*(subvolDim-2); dst_off.y = 1 + _y*(subvolDim-2); dst_off.z = 1 + _z*(subvolDim-2);
 	copy3DMemToHost(d_volPPtr, h_phi, copyvol, volSize, src_off, dst_off); 
     	CHK_CUDA_ERR("->");
      }
   
  CUT_SAFE_CALL(cutStopTimer(timer));
  float milliseconds = cutGetAverageTimerValue(timer);
  fprintf(stderr, " %f ms\n", milliseconds);
  CUT_SAFE_CALL(cutResetTimer(timer));

  return true;
}

int main(int argc, char **argv)
{
  CUT_DEVICE_INIT(argc, argv);
  CUT_SAFE_CALL(cutCreateTimer(&timer));
  CUT_SAFE_CALL(cutResetTimer(timer));

  //load volume data
  h_phi = loadRawData();
  if(h_phi == NULL) {fprintf(stderr, "Can't load 3D array.\n"); return 1;}
  
  // Allocate cuda array and global mem for subvol. Bind texture to cuda array
  initCuda();

  initInterface();// Initialize level set interface for the phi function

  writeRawData(h_phi, "raw.out");
  free(h_phi);
  CUDA_SAFE_CALL(cudaFreeArray(d_volArray));
  CUDA_SAFE_CALL(cudaFree(d_volPPtr.ptr));
  CUT_SAFE_CALL(cutDeleteTimer(timer));

  return EXIT_SUCCESS;

}

